;**********************************************************************
;
;    Filename:	    pbus628.asm
;    Description:   rs485-type multi-drop bus
;                   with half-duplex serial protocol
;                   one master, max 15 slaves
;                   inv. logic: 0V=+12V=bit0 (START)
;                                 5V=-12V=bit1 (STOP)
;
;    Author:        el@jap.hu
;                   http://jap.hu/electronic/
;
;**********************************************************************
; HISTORY
;
;060 - first hardware USART version for the 16f628
;061 - working receive & response, only reinit need a separate timer itr
;062 - 20010623
;063 - 20010623
;064 - 20010623 - working reinit alg! adjust timers, remove DEBUG lines
;065 - 20010624 - timers adjusted - fine-tune maxdelay!
;066 - 20010624 - first cleanup
;067 - 20010624 - add TX delay
;068 - 20010624 - optional KEEP_STATS
;069 - 20090816 - dynamic devid; export txbuf, rxbuf for pbus_handler
;**********************************************************************
;
;    Notes:
;
; To use the interrupt handler, SAVE these registers:
; W, STATUS, FSR (if used in main program!)
; TMR1 is reserved for RX/TX scheduling
;
;**********************************************************************

#define KEEP_STATS

	GLOBAL pbus_init, pbus_int_handler
	GLOBAL rxbuf, txbuf
#ifdef KEEP_STATS
	GLOBAL rxe1cnt, rxe2cnt, rxe3cnt
#endif
	EXTERN pbus_handler

	list      p=16f628

#include <p16F628.inc>

#define ROK 0x60
#define RECHO 0x6f

#define DBIT PORTB, 3

.pbusda	UDATA
;***** VARIABLE DEFINITIONS

GLOBAL bytecnt
bytecnt	res 1 ; serial routine flags
chksum	res 1 ; buffer checksum
devid	res 1 ; device ID

#ifdef KEEP_STATS
rxe1cnt	res 2 ; rx checksum error counter
rxe2cnt	res 2 ; rx all packets counter
rxe3cnt	res 2 ; rx my packets counter
#endif

; don't change buffer order!
rxbuf	res 8 ; receive buffer
txbuf	res 8 ; transmit buffer
rmax	equ 8
tmax	equ 8

; * CONSTANTS

#define PBUS_VER 0x69
#define MAIN_VER 0x10

bittime		equ .417
rxdelay		equ .10 * bittime	; delay time to switch from last byte TX to RX
txdelay		equ .1 * bittime	; delay time to switch from last byte RX to TX
maxdelay	equ .4900		; max delay in packet between bytes
					; rxdelay: STOP + START + 8data
					; txdelay: starts at the end of receive STOP

t1prescaler	equ .8
t1rxdelay	equ .65536 - (rxdelay / t1prescaler)
t1txdelay	equ .65536 - (txdelay / t1prescaler)
t1maxdelay	equ .65536 - (maxdelay/ t1prescaler)

f_osc		equ .4000000
ser_baudrate	equ .2400
ser_baud	equ (f_osc/(ser_baudrate*.16)) - 1

.pbuslib	CODE

pbus_cmd_vectors
		addlw pbus_cmd
		movwf PCL
pbus_cmd	; command jump table
		goto cnoop		; 8: noop
		goto rxb_ena		; 9: illegal
		goto rxb_ena		; a: illegal
		goto clast		; b: clast

#ifdef KEEP_STATS
		goto cres		; c: cres
		goto cstat		; d: cstat
#else
		goto rxb_ena		; c: illegal
		goto rxb_ena		; d: illegal
#endif

		goto cver		; e: cver
		goto cping		; f: cping

pbus_init	;
		movwf devid
		clrf bytecnt

#ifdef KEEP_STATS
		clrf rxe1cnt
		clrf rxe1cnt+1
		clrf rxe2cnt
		clrf rxe2cnt+1
		clrf rxe3cnt
		clrf rxe3cnt+1
#endif

		movlw rxbuf
		movwf FSR ; init rxbuf and enable rx
		clrf chksum

		; init serial, enable RX itr
		BANKSEL SPBRG
		movlw ser_baud
		movwf SPBRG    ;baud rate register
		clrf TXSTA

		movlw 0x24	; +BRGH, +TXEN, -SYNC, -TX9
		movwf TXSTA
		bsf PIE1, RCIE  ;enable interrupt on rx
		bcf PIE1, TXIE	;disable interrupt on tx
		bcf TRISB, 2	;tx is output
		bsf TRISB, 1	;rx input
		BANKSEL RCSTA
		clrf RCSTA
		bsf RCSTA, SPEN	;pins for serial
		bsf RCSTA, CREN	;enable rx

		; init TMR1
		movlw 0x31	; -TMR1CS, +T1CON<4:5>, -TMR1ON
		movwf T1CON

		BANKSEL PIE1
		bsf PIE1, TMR1IE ; enable TMR1 interrupt
		BANKSEL PIR1

		movlw low t1maxdelay
		movwf TMR1L
		movlw high t1maxdelay
		movwf TMR1H
		bsf T1CON, TMR1ON ; enable TMR1

		movlw (1<<GIE)+(1<<PEIE)
		movwf INTCON
		return

pbus_int_handler; interrupt handler

		btfsc PIR1, TMR1IF
		goto tmrint

		btfsc PIR1,RCIF
		call rxint

		btfss PIR1, TXIF
		return

		btfsc RCSTA, CREN
		return ; in RX mode!

txint		movf bytecnt, W
		bz txbdone

		comf chksum, W	; last byte in packet is chksum
		addlw 1
		decfsz bytecnt, F
		movf INDF, W	; transmit a byte from packet

		movwf TXREG
		addwf chksum, F
		incf FSR, F
		return


txbdone		; TX done
		; automatically schedule a TX-> RX switch - rxdelay

		BANKSEL PIE1
		bcf PIE1, TXIE	; disable TX itr:TXIF won't go away - endless loop
		BANKSEL txbuf

		bcf T1CON, TMR1ON
		bcf PIR1, TMR1IF
		movlw low t1rxdelay
		movwf TMR1L
		movlw high t1rxdelay
		movwf TMR1H
		bsf T1CON, TMR1ON ; enable TMR1
		return

tmrint		; timer int
		; decide if a receive timeout/switch (goto rxb_ena)
		; or a transmit switch

		bcf PIR1, TMR1IF  ; clear TMR1 itr flag

		movlw txbuf
		subwf FSR, W
		bz sched_tx

rxb_ena
		bcf DBIT

sched_rx	bsf RCSTA, CREN ; enable receiver
		bcf T1CON, TMR1ON ; disable TMR1

		movlw rxbuf
		movwf FSR ; init rxbuf and enable rx
		clrf bytecnt

		bsf DBIT
		return

sched_tx	; turn TX on
		bcf DBIT

		bcf RCSTA, CREN ; disable RX
		bcf T1CON, TMR1ON ; disable TMR1
		BANKSEL PIE1
		bsf PIE1, TXIE	; schedule TX itr
		BANKSEL txbuf
		return

rxint		; RX: byte received itr

		; a byte was received, set timeout again
		bcf T1CON, TMR1ON ; disable TMR1
		movlw low t1maxdelay
		movwf TMR1L
		movlw high t1maxdelay
		movwf TMR1H
		bsf T1CON, TMR1ON ; enable TMR1

		movf bytecnt, W
		bnz rx_2

		; first byte
		movf RCREG, W
		movwf rxbuf
		movwf chksum
		movlw 0x0f
		andwf rxbuf, W
		addlw 3
		movwf bytecnt
		goto rx_3

rx_2		movf RCREG, W
		movwf INDF
		addwf chksum, F

rx_3		incf FSR, F
		decfsz bytecnt, F
		return

rxb_done	; packed received

		movf chksum, W		; check packet
		btfsc STATUS, Z
		goto skip_6

#ifdef KEEP_STATS
		incf rxe1cnt, F		; packet checksum error
		btfsc STATUS, Z
		incf rxe1cnt+1, F
#endif
		goto rxb_ena		; restart with new packet

skip_6

#ifdef KEEP_STATS
		incf rxe2cnt, F		; count good pakets
		btfsc STATUS, Z
		incf rxe2cnt+1, F
#endif

		movlw 0xf0		; check devid
		andwf rxbuf, W
		subwf devid, W
		btfss STATUS, Z
		goto rxb_ena		; devid mismatch, restart

;************** Command handlers start here
; illegal commands with no response goto rxb_ena (restart with a new packet)
; good commands with resp. in txbuf goto txb_ena (TX a packet with checksum)
;**************
handle_cmd	; got a good packet

#ifdef KEEP_STATS
		incf rxe3cnt, F
		btfsc STATUS, Z
		incf rxe3cnt+1, F
#endif

		movlw 0xf8		; check for higher 5 bits
		andwf (rxbuf+1), W
		sublw 0x58
		btfss STATUS, Z
		goto cext		; not internal

		movlw 0x07
		andwf (rxbuf+1), W
		goto pbus_cmd_vectors

cping		movlw RECHO
		movwf txbuf+1

		movlw 0x0f
		andwf rxbuf, W
		movwf txbuf
		movwf chksum		; use as counter
		bz txb_ena		; data present?

		movlw rxbuf+2		; copy data to response
		movwf FSR

cpingcopy	movf INDF, W
		movwf rxbuf+1		; use as temp storage
		movlw rmax
		addwf FSR, F		; goto transmit buffer
		movf rxbuf+1, W
		movwf INDF

		movlw rmax
		subwf FSR, F		; goto receive buffer
		incf FSR, F
		decfsz chksum, F
		goto cpingcopy

		goto txb_ena

cver		movlw 2
		movwf txbuf
		movlw ROK
		movwf txbuf+1
		movlw PBUS_VER
		movwf txbuf+2
		movlw MAIN_VER
		movwf txbuf+3
		goto txb_ena

#ifdef KEEP_STATS
cstat		movlw 6
		movwf txbuf
		movlw ROK
		movwf txbuf+1
		movf rxe1cnt, W
		movwf txbuf+2
		movf rxe1cnt+1, W
		movwf txbuf+3
		movf rxe2cnt, W
		movwf txbuf+4
		movf rxe2cnt+1, W
		movwf txbuf+5
		movf rxe3cnt, W
		movwf txbuf+6
		movf rxe3cnt+1, W
		movwf txbuf+7
		goto txb_ena

cres		clrf rxe1cnt
		clrf rxe1cnt+1
		clrf rxe2cnt
		clrf rxe2cnt+1
		clrf rxe3cnt
		clrf rxe3cnt+1
#endif

cnoop		movlw 0
		movwf txbuf
		movlw ROK
		movwf rxbuf+1
		goto txb_ena

cext		movf (rxbuf+1), W
		call pbus_handler
		andlw 0xff		; if at return W=0, respond
		btfss STATUS, Z
		goto rxb_ena		; otherwise, go back listening

clast		; repeat last response (in the transmit buffer)

txb_ena		; schedule a RX->TX switch after RXB to answer - txdelay

		bcf T1CON, TMR1ON ; disable TMR1
		movlw low t1txdelay
		movwf TMR1L
		movlw high t1txdelay
		movwf TMR1H
		bsf T1CON, TMR1ON

		movlw txbuf
		movwf FSR
		clrf chksum
		movlw 0x0f
		andwf txbuf, W
		addlw 0x03
		movwf bytecnt
		return

	end

